package multex; //Exc.java
import java.lang.String;

//History:
//2000-10-06  Knabe  Require default message text pattern in the exception
//2000-09-29  Knabe  Allows only named parameters
//2000-09-28  Knabe  _checkClass checks public-ness of the class, too
//2000-07-31  Knabe  getMessage() formats parameter as {i}=.. rather than [i]=..
//2000-06-29  Knabe  _cause gets type Exception rather than Throwable
//2000-06-19  Knabe  printStackTrace() bei Ursachenkette ohne Ausgabe
//                   von redundanten Zeilen.
//2000-06-12  Knabe  alle printStackTrace(...) ber printStackTrace(PrintWriter)
//1999-03-23  Knabe  Parametrierung wie in Java-Tutorial/i18n/format
//1999-03-12  Knabe  NEU: causingThrowable()
//1999-01-14  Knabe  erstellt

/**Problem-related exception with named parameters and internationalizable
  message text.
  This class serves for the framework user as a base class for defining
  problem-related exceptions.

  <H2>Naming convention:</H2>
  All user-defined exception classes derived from Exc should have a name
  ending in Exc. The pattern is: <PRE>
    class ErrorconditionExc extends Exc { ... }
  </PRE>

  <H2>Named parameters</H2>
  All public non-static fields of an exception object are
  considered named parameters of it. The corresponding placeholders in the
  message text patterns have the syntax: {name}
  with name being the name of the parameter.
  <P>
  This way of providing parameters is the recommended one for the following
  reasons: Firstly it is
  easier to translate a message text pattern into another language, if in the
  message text pattern there is an indication, for which purpose the
  message parameter is. The parameter name usually is a good hint to this.
  Secondly naming parameters is safer when adding or removing a parameter
  to/from an exception.
  Thirdly named parameters, for which a placeholder in the corresponding
  message text is missing, will not be suppressed. They will be appended
  in a standard format along with their name to the end of the message text.
  <P>
  The exception parameters can
  be substituted into the corresponding message text pattern in a desired locale
  format by class {@link MsgText}, which itself makes use of class
  java.text.MessageFormat.
  <br>
  See the usage examples Copy, AwtCopy, SwingCopy in directory demo.

  <H2>Relation to Standards</H2>
  Class java.text.MesageFormat expects the message parameters
  to be stored in an Object[]. So they are indexed parameters.
  We did not follow this way in order to be able to name parameters.
  All the rest of parameter formatting is done by and like
  java.text.MesageFormat, e.g. formatting a parameter of type java.util.Date
  with name today could be done by the following parameter formatting directive:
  {today,date}.

  @author Christoph Knabe, TFH Berlin, 1999-2000 Copyrighted
*/

public abstract class Exc extends java.lang.Exception
implements MultexException {


private static String _className = "Exc";

/** Constructs an Exc with a default message text pattern.
  All public nonstatic fields are considered as named
  parameters of this exception object.
*/
public Exc(final String i_defaultMessageTextPattern){
  _defaultMessageTextPattern = i_defaultMessageTextPattern;
  _checkClass();
}

/**
  Returns the exception parameters of this object or null.
  They are in a compact, human readable format, e.g.:
  <PRE>
    {name}=myfile.dat  {directory}=C:\temp\data
  </PRE>
*/
public String getMessage() {
  final java.lang.reflect.Field[] namedParams = getPublicNostaticFields(this);
  if(namedParams.length==0){return null;}
  final StringBuffer result = new StringBuffer();
  appendNamedParams(result, namedParams, this);
  return result.toString();
}

/**Appends the name/value pairs of all public fields of i_exception,
  which are listed in i_params to io_buffer. The notation is:
  {name1}=value1  {name2}=value2
*/
/*package*/ static void appendNamedParams(
  final StringBuffer io_buffer,
  final java.lang.reflect.Field[] i_params,
  final Exception i_exception
){
  for(int i=0; i<i_params.length; i++){
    final java.lang.reflect.Field field = i_params[i];
    Object fieldValue;
    try{fieldValue = field.get(i_exception);}
    catch(IllegalAccessException ex){fieldValue = "<<inaccessible>>";}
    appendParam(io_buffer, field.getName(), fieldValue);
  }
}

public String getDefaultMessageTextPattern(){
  return _defaultMessageTextPattern;
}

/**Checks that the class of this object is OK*/
private void _checkClass(){
  checkClass(this, _className);
}

/**Reports to System.err, if the class of this exception object
does not satisfy the conditions: <UL>
  <LI> end with i_suffix
  <LI> is public
</UL>
This checking should better occur statically before running a program
and not dynamically each time an exception object is created.
*/
/*package*/ static void checkClass(final Exception i_exception, final String i_suffix){
  //check, that class name ends with i_suffix:
  final Class exceptionClass = i_exception.getClass();
  final String exceptionName = exceptionClass.getName();
  if(!exceptionName.endsWith(i_suffix)){
    System.err.println("The name of subclass "+ exceptionName +" of " + i_suffix
      + " should end in " + i_suffix + "!"
    );
  }
  //check, that the class is public in order to access its attributes:
  final int modifiers = exceptionClass.getModifiers();
  if(!java.lang.reflect.Modifier.isPublic(modifiers)){
    System.err.println(exceptionName + ": This class must be public for exception parameter access!");
  }
}

/**Appends to io_buf the external representation "{i_name}=i_value  "
*/
/*package*/ static void appendParam(
  final StringBuffer io_buf, final String i_name, final Object i_value
){
  io_buf.append("{");   io_buf.append(i_name);   io_buf.append("}=");
  io_buf.append(i_value);   io_buf.append("  ");
}

/**Returns the Reflections representation of all public fields in i_object*/
/*package*/ static java.lang.reflect.Field[]
getPublicNostaticFields(final Object i_object){
  final Class theClass = i_object.getClass();
  final String className = theClass.getName();
  final java.lang.reflect.Field[] publicFields = theClass.getFields();
  int statics = 0;
  for(int i=0; i<publicFields.length; i++){ //check no static fields
    final java.lang.reflect.Field field = publicFields[i];
    final int modifiers = field.getModifiers();
    if(java.lang.reflect.Modifier.isStatic(modifiers)){
      publicFields[i] = null;   statics++;
    }
  }
  //Create result with only the nonstatic fields
  final java.lang.reflect.Field[] result
  = new java.lang.reflect.Field[publicFields.length-statics];
  for(int zi=0,qi=0; qi<publicFields.length; qi++){
    final java.lang.reflect.Field field = publicFields[qi];
    if(field!=null){result[zi++] = field;}
  }
  return result;
}


private /*Java1.2:final*/ String _defaultMessageTextPattern;


}//Exc