package multex; //Failure.java

//History:
//2000-10-06  Knabe  Require default message text pattern in the exception
//2000-07-13  Knabe  Efficiently reimplemented: printCompactStackTrace
//2000-07-09  Knabe  Retrofitted to JDK 1.1: Uses Vector instead of List
//2000-06-29  Knabe  Class Exc split into Exc and Failure


/**Indirectly caused exception with parameters, causal chain and message text.
  This class serves for the framework user as a base class for defining
  Failure-exceptions, which have the following benefits: <UL>
    <LI> coherent usage always for unexpected failure of an operation
    <LI> reduce the amount of exceptions to specify in the throws-clause
         from many unexpected exceptions to only one
    <LI> chance to provide the failure of an operation with a message text,
         message parameters and a chain of the unexpected causing exceptions.
  </UL>
  The named exception parameters are handled the same way as in class {@link Exc}.
  For convenience ist is additionally possible to give unnamed parameters
  to this exception. But they cannot be inserted into a message text pattern.
  They will be only appended to the text in a standard way for diagnostics
  output.
  <P>
  Naming convention:
    All user-defined exception classes derived from Failure should have a name
    ending in Failure. The pattern is: <PRE>
      class OperationnameFailure extends Failure { ... }
    </PRE>
  <H2>Failure as unchecked exception?</H2>
  It can be bothering always to specify Failure in the throws-clause
  of any method, that calls other non-trivial methods. This holds especially,
  when you are converting an existing software system to usage of MulTEx.
  Thus Failure inherits the unchecked java.lang.RuntimeException.
  <P>
  So you can specify Failure, but you need not to do this.
  This is in my opinion not the cleanest way, but even without Failure
  you could not base on the assumption, that a method without a throws-clause
  does not throw any exception. Any method could throw any
  java.lang.RuntimeException, e.g. the IndexOutOfBoundsException.
  <P>
  If you want to strictly check Failure, then you can
  change the superclass of Failure to the checked java.lang.Exception instead of
  the unchecked java.lang.RuntimeException. Up to now I am not sure, what is the
  best way. Please report your experiences.
  @author Christoph Knabe, knabe@tfh-berlin.de, TFH Berlin, 1999-2000 Copyrighted
*/

public abstract class Failure extends RuntimeException
implements MultexException {


private static String _className = "Failure";

/**The string used to delimit the stack trace part of a causing Exception from
  the following stack trace of the caused Failure-exception.
*/
private static final String _wasCausing = "WAS CAUSING:";

/** Constructs a parameterized Failure-exception, giving only the causing Exception object.
* All public nonstatic attributes of this Failure-exception are considered as
* named exception parameters and can be substituted into the corresponding
* message text. This is the preferred way to give parameters to an exception,
* as the meaning of the parameters is more self-documenting.
* <P>Example of defining an exception with cause and the named parameters
* {inFileName} and {outFileName}: <PRE>
*   public static class CopyFailure extends Failure {
*     public final String inFileName;
*     public final String outFileName;
*     public CopyFailure(
*       final Exception i_cause,
*       final String    i_inFileName,
*       final String    i_outFileName
*     ){
*       super("File '{inFileName}' could not be copied to '{outFileName}'",i_cause);
*       inFileName = i_inFileName;
*       outFileName = i_outFileName;
*     }
*   }//CopyFailure
* </PRE>
*/
public Failure(
  final String i_defaultMessageTextPattern, final Exception i_cause
){
  this(i_defaultMessageTextPattern, i_cause, (Object[])null);
}

/**Constructs a Failure-exception, giving the causing Exception object for
  providing the diagnostics causer chain and unnamed message parameters as an
  Object[].
  The preferred way to parameterize a Failure exception is by giving named
  parameters to a subclass of Failure,
  which is described at {@link #Failure(String,Exception)}
  In order to shortly create unnamed message parameters, you could use
  the convenience class {@link MethodFailure}, which passes its unnamed
  parameters to the polymorphic Object[] i_unnamedParameters.
*/
public Failure(
  final String i_defaultMessageTextPattern,
  final Exception i_cause, final Object[] i_unnamedParameters
){
  _defaultMessageTextPattern = i_defaultMessageTextPattern;
  _cause = i_cause;
  _unnamedParameters = i_unnamedParameters;
  _checkClass();
}

public String getDefaultMessageTextPattern(){
  return _defaultMessageTextPattern;
}

/**Result: The Exception, which caused the program to throw this
  Failure-exception. I.e. the Exception, which was redefined into this Failure
  object.
*/
public final Exception cause(){return _cause;}

/** Result: the causing Exception of i_throwable, if there is one,
* otherwise null.
*/
public final static Exception cause(final Throwable i_throwable){
  return i_throwable instanceof Failure
    ? ((Failure)i_throwable).cause()
    : null
  ;
}

/**
  Returns the exception parameters of this object or null.
  Firstly come the named parameters, then the unnamed ones.
  They are in a compact, human readable format, e.g.:
  <PRE>
    {name}=myfile.dat  {directory}=C:\temp\data  {0}=3353  {1}=rw
  </PRE>
*/
public String getMessage() {
  final java.lang.reflect.Field[] namedParams = Exc.getPublicNostaticFields(this);
  final boolean noIndexParams = _unnamedParameters==null;
  final boolean noNamedParams = namedParams.length==0;
  if(noIndexParams&&noNamedParams){return null;}

  final StringBuffer result = new StringBuffer();
  Exc.appendNamedParams(result, namedParams, this);
  appendUnnamedParams(result);
  return result.toString();
}

/**Appends the index/value pairs of all unnamed parameters of this Failure.
  The notation is: <PRE>
  {0}=value0  {1}=value1  {2}=value2
  </PRE>
*/
/*package*/ void appendUnnamedParams(
  final StringBuffer io_buffer
){
  if(_unnamedParameters==null){return;}

  for(int i=0; i<_unnamedParameters.length; i++){
    Exc.appendParam(io_buffer, Integer.toString(i), _unnamedParameters[i]);
  }
}

/** Prints the chained, redundant stack traces of this <code>Failure</code>
  object to the standard error stream System.err
  @see #printStackTrace(java.io.PrintWriter)
  @see java.lang.System#err
*/
public void printStackTrace(){
  printStackTrace(System.err);
}

/** Prints the chained, redundant stack traces of this <code>Failure</code>
  object to io_printer.
* @see #printStackTrace(java.io.PrintWriter)
*/
public void printStackTrace(final java.io.PrintStream io_printer) {
  printStackTrace(new java.io.PrintWriter(io_printer, true));
}

/** Prints the chained, redundant stack traces of this <code>Failure</code>
  object to io_printer.
  The output contains the name, message parameters and stack trace first of
  the causing <code>Exception</code> object, then of this <code>Failure</code>
  object to the specified print writer. Users should usually not call this
  method directly, but {@link Msg#printStackTrace} instead, in order to avoid
  redundant and useless lines in the stack traces.
*/
public void printStackTrace(final java.io.PrintWriter io_printer) {
  //simple efficient code: leaves all redundant lines:
  if(_cause!=null){
    _cause.printStackTrace(io_printer);
    io_printer.println(_wasCausing);
  }
  super.printStackTrace(io_printer);
  io_printer.flush();
}


/** Prints the stack trace of i_throwable and all its chained cause exceptions
  to io_printer. All redundant location lines therein are suppressed.
  As well all location lines, which are also contained in the stack trace of
  i_here, are suppressed.
  This is useful, e.g. for reporting an error in a Swing-ActionListener,
  because the stack trace of an actionPerformed-method contains 19 lines
  inside of the AWT/Swing-Package. These lines are not useful for error
  searching, as they are always the same in any Swing application.
  Usually you should not directly call this method, but a report-method in
  class Msg.
  @param io_printer Where to print the compactified stack trace
  @param i_reportee The exception to report, possibly containing a chain of
    nested cause exceptions
  @param i_here An dummy exception, which should have been newly created there,
    where the reporting for i_reportee is invoked
  @see Msg#report(java.awt.Frame,Throwable)
*/
public static void printCompactStackTrace(
  final StringBuffer io_buffer, final Throwable i_reportee,
  final Exception i_here
){
  //prepare out writer:
  final java.io.StringWriter sw = new java.io.StringWriter();
  final java.io.PrintWriter out = new java.io.PrintWriter(sw, true);
  //final java.io.PrintWriter err = new java.io.PrintWriter(System.err, true);

  //collect stack traces onto out writer:
  i_reportee.printStackTrace(out);
  out.println(_wasCausing);
  i_here.printStackTrace(out);
  final String tracesString = sw.toString();
  //System.err.println("tracesString:");System.err.println(tracesString);//debug
  final java.util./*String*/Vector tracesLines
  = _splitIntoLines(tracesString);

  //nullify redundant trace trail lines:
  final int[] traceEndIndices = _getTraceEndIndices(tracesLines);
  //_println("traceEndIndices", traceEndIndices); //debug
  for(int i=1; i<traceEndIndices.length; i++){
    _nullifyCommonTraceTail(
      tracesLines, traceEndIndices[i-1], traceEndIndices[i]
    );
  }
  _shortenLastTrace(
    tracesLines, traceEndIndices[traceEndIndices.length-2]+1
  );
  _printLines(io_buffer, tracesLines);
}//printCompactStackTrace

/**Splits i_string into internized line strings by BufferedReader.readln().
  @return The lines in the sequence of storage in i_string as a Vector of
          String-s.
*/
private static java.util./*String*/Vector _splitIntoLines(final String i_string){
  final java.io.BufferedReader reader = new java.io.BufferedReader(
    new java.io.StringReader(i_string)
  );
  final java.util./*String*/Vector result = new java.util.Vector();
  for(;;){
    String line; //Warum geht es hier nicht final?
    try{line = reader.readLine();}
      catch(java.io.IOException ex){line=null;}
    //end try
    if(line==null){break;}
    result.addElement(line.intern());
  }
  return result;
}

/**Result: The indices of the elements of i_tracesLines, which are immediately
  before the occurences of the String _wasCausing in i_tracesLines.
  Require: In i_traces are several stack traces, split into lines; the traces
    are separated by the String _wasCausing.
*/
private static int[] _getTraceEndIndices(final java.util.Vector i_tracesLines){
  final int numberOfTraces
  = _numberOfOccurences(i_tracesLines, _wasCausing) + 1;
  final int[] result = new int[numberOfTraces];
  final int size = i_tracesLines.size();
  int resultIndex = 0;
  for(int i=0; i<size; i++){
    if(i_tracesLines.elementAt(i) == _wasCausing){
      result[resultIndex++] = i-1; //line before _wasCausing
    }
  }
  result[resultIndex++] = size-1; //last line
  return result;
}

/**Result: The number, how often i_object occurs in i_vector,
  measured by the identity comparison
*/
private static int _numberOfOccurences
(final java.util.Vector i_vector, Object i_object){
  final int size = i_vector.size();
  int result = 0;
  for(int i=0; i<size; i++){
    if(i_vector.elementAt(i) == i_object){result++;}
  }
  return result;
}

/**Nullifies redundant entries in io_vector.
  Nullifies all entries in io_vector
  starting from i_nullifyEndIndex inclusively downwards,
  which are identically contained in the same
  sequence as entries starting from i_compareEndIndex inclusively downwards.
*/
private static void _nullifyCommonTraceTail(
  final java.util.Vector io_vector,
  final int i_nullifyEndIndex, final int i_compareEndIndex
){
  for(int ni=i_nullifyEndIndex, ci=i_compareEndIndex; ;ni--,ci--){
    if(io_vector.elementAt(ni)!=io_vector.elementAt(ci)){break;}
    io_vector.setElementAt(null, ni);
  }
}//_nullifyCommonTraceTail

/**Shortens the Vector io_tracesLines by setting the element at index
  i_fromIndex to the abbreviation symbol "at ..." and nullifying all
  elements after this.
*/
private static void _shortenLastTrace(
  final java.util.Vector io_tracesLines, final int i_fromIndex
){
  io_tracesLines.setElementAt("\tat ...", i_fromIndex);
  final int size = io_tracesLines.size();
  for(int i=i_fromIndex+1; i<size; i++){
    io_tracesLines.setElementAt(null, i);
  }
}

/**Appends the non-null elements of i_lines each as a line to io_buffer*/
private static void _printLines(
  final StringBuffer    io_buffer,
  final java.util.Vector i_lines
){
  final int size = i_lines.size();
  for(int i=0; i<size; i++){
    final Object line = i_lines.elementAt(i);
    if(line!=null){io_buffer.append(line);io_buffer.append(MsgText.lineSeparator);}
  }
}

/**Checks that the class of this object is OK*/
private void _checkClass(){
  Exc.checkClass(this, _className);
}

/**Result:
  Preferably use named parameters (public non-static fields) for individual
  Failure-exceptions derived from Failure.
*/
/*package*/ final Object[] unnamedParameters(){return _unnamedParameters;}

private /*Java1.2:final*/ String _defaultMessageTextPattern;

/**The Exception object, that caused this Failure-exception
  @see #cause()
*/
private /*Java1.2:final*/ Exception _cause;

/**Additional unnamed parameters of this exception.*/
private /*Java1.2:final*/ Object[] _unnamedParameters;



}//Failure