package multex; //MsgText.java



//nderungshistorie:
//2000-09-29  Knabe  Unnamed parameters are no longer substituted into a
//                   message text, only named ones (public nostatic fields)
//2000-07-24  Knabe  Umsetzung von Namensparametern mittels Reflection
//2000-07-07  Knabe  extrahiert aus Klasse Msg


/** Services for getting the message text for exceptions.
  It handles exceptions <UL>
    <LI>with parameters (subclasses of {@link Exc})
    <LI>with cause chain and parameters (subclasses of {@link Failure})
    <LI>and standard Java exceptions (other subclasses of java.lang.Throwable)
  </UL>
  @author Christoph Knabe, Berlin, Copyright 1999-2000
*/
public class MsgText {


private static final String _className = "MsgText";
private static final String _because = "CAUSE: ";

/**Returns the message for i_throwable with its causal chain.
  @return A usually multi-line String, in which the first line is the
    message for i_throwable and the further lines are the messages in the
    chain of causing exceptions of i_throwable. The lines are separated by
    {@link #lineSeparator}. See {@link #getMessageLine} for a despription,
    how each exception is transformed into a message line.
*/
public static String getMessageChain(final Throwable i_throwable){
  final StringBuffer result = new StringBuffer(160);
  //Throwable-Meldungen eintragen:
  for(Throwable t=i_throwable; t!=null; t=Failure.cause(t)){
    if(t!=i_throwable){ //causing exception
      result.append(lineSeparator);
      result.append(_because);
    }
    result.append(getMessageLine(t));
  }
  return result.toString();
}


/**Result: A usually one-line message text for the exception
  i_throwable with inserted parameters.

  The message text for the exception is taken from the properties file
  "MsgText.properties" or one of the localized variants
  of it, e.g. "MsgText_de_DE.properties", see the Java Tutorial
  <A HREF="http://java.sun.com/docs/books/tutorial/i18n/format/messageFormat.html">
  chapter on internationalization</A> at february 1999.
  <P>
  This file must contain definition lines of the form: <PRE>
     packageName.ExceptionName = message text pattern
  </PRE>or<PRE>
     packageName.ClassName$InnerExceptionName = message text pattern
  </PRE>with parameter substitutors in the message text pattern, e.g. with named parameters:<PRE>
     CopyNP$CopyFailure = The file "{inFileName}" could not be copied to "{outFileName}"
  </PRE>or with (deprecated) indexed parameters:<PRE>
     Copy$CopyFailure = The file "{0}" could not be copied to "{1}"
  </PRE>If i_throwable is a standard Java exception with one String as
  detailMessage, this is taken as message parameter {0}.
*/
public static String getMessageLine(final Throwable i_throwable){
  /*final String className = i_throwable.getClass().getName();
  _mut_tryLoadMessages();
  if(_messages==null){
    return i_throwable.toString();
  }*/
  try {
    final StringBuffer result = new StringBuffer();
    final String namedParamsText = _getMessageTextPattern(i_throwable);
    /*
    try{namedParamsText = _messages.getString(className);}
      catch(java.util.MissingResourceException ex){
        return i_throwable.toString();
      }
    //end try*/
    final _Param[] namedParams = _getParams(i_throwable);
    final String formattedMessage;
    if(namedParams==null){ //no insertable exception parameters
      formattedMessage = namedParamsText;
    }else{ //with named exception parameters:
      //transform into index parameters:
      final String indexedParamsText
      = _replaceNameByIndexPlaceholders(namedParamsText,namedParams);
      final Object[] indexedParams
      = _distributeParams(result, namedParams);
      formattedMessage = java.text.MessageFormat.format(
        indexedParamsText, indexedParams
      );
    }
    result.insert(0, formattedMessage);
    if(i_throwable instanceof Failure){ //may have unnamed parameters:
      final Failure failure = (Failure)i_throwable;
      failure.appendUnnamedParams(result);
    }else if(i_throwable instanceof Exc){ //cannot have unnamed parameters:
    }else{ //other Java exceptions may have detailMessage:
      final String detailMessage = i_throwable.getMessage();
      if(detailMessage!=null){
        //Append standard Java detailMessage, separated by a colon
        result.append(": ");
        result.append(detailMessage);
      }
    }
    return result.toString();
  } catch(final Exception ex){
      System.err.println(_className + ": EXCEPTION OCCURRED WHEN PREPARING MESSAGE LINE");
      ex.printStackTrace();
      return i_throwable.toString();
    }
  //end try
}//getMessageLine

/**Gets a message text pattern for the exception i_throwable.
  The following sources are tried one after another with the class name
  of i_throwable as message text key: <OL>
    <LI>The language and country specific MsgText properties file,
        e.g. MsgText_en_UK.properties
    <LI>The language specific MsgText properties file,
        e.g. MsgText_en.properties
    <LI>The default MsgText properties file: MsgText.properties
    <LI>The default message text pattern from i_throwable, if it is a
        MultexException
    <LI>The class name of the exception i_throwable
  </OL>
*/
private static String _getMessageTextPattern(final Throwable i_throwable){
  final String className = i_throwable.getClass().getName();
  final String pattern0 = i_throwable instanceof MultexException ?
    ((MultexException)i_throwable).getDefaultMessageTextPattern() : className
  ;
  final String defaultPattern = pattern0==null ? className : pattern0;
  _mut_tryLoadMessages();
  if(_messages==null){
    return defaultPattern;
  }
  try{return _messages.getString(className);}
    catch(java.util.MissingResourceException ex){
      return defaultPattern;
    }
  //try
}

/**Returns the default message text pattern for i_throwable or null.
  Inserting the exception parameters into the message text pattern is
  done by class {@link MsgText}.
  Locale specific message text patterns are handled by MsgText, too.
private static String __getDefaultTextPattern(final Throwable i_throwable){
  try{
    final java.lang.reflect.Field field
    = i_throwable.getClass().getField("text");
    final Object text = field.get(i_throwable);
    return text.toString();
  } catch(Exception ex){return null;}
}
*/

/**Returns the named parameters contained in i_throwable*/
private static _Param[] _getParams(final Throwable i_throwable)
throws IllegalAccessException {
  final java.lang.reflect.Field[] namedParams
  = Exc.getPublicNostaticFields(i_throwable);
  if(namedParams.length<=0){return null;}
  //Copy namedParams contents:
  final _Param[] result = new _Param[namedParams.length];
  for(int i=0; i<namedParams.length; i++){
    final java.lang.reflect.Field source = namedParams[i];
    final _Param dest
    = new _Param(source.getName(), source.get(i_throwable));
    result[i] = dest;
  }
  return result;
}

/**Returns i_messagePattern, in which the name-placeholders
  are replaced by index-placeholders. Only placeholders for the names of the
  parameters in io_params are replaced by the placeholder for the corresponding
  index. Parameters used in this way are marked as inText.
*/
private static String _replaceNameByIndexPlaceholders(
  final String i_messagePattern,
  final _Param[] io_params
){
  final StringBuffer result = new StringBuffer(i_messagePattern.length());
  for(int ci=0; ci<i_messagePattern.length(); ci++){ //char loop
    final char c = i_messagePattern.charAt(ci);
    result.append(c);
    if(c=='{'){ //may be start of name parameter placeholder
      final int pi = _indexForMatchedParam(i_messagePattern, io_params, ci+1);
      if(pi!=_noMatch){ //placeholder found in message text pattern:
        //???Needs still translate the supress syntax. e.g.: {name-}
        result.append(pi);
        //skip param name:
        ci += io_params[pi].name.length();
        io_params[pi].inText = true;
      }//if pi
    }//if c
  }//for ci
  //System.err.println("_replaceNameByIndexPlaceholders-Result: " + result);
  //_println("_replaceNameByIndexPlaceholders-io_params", io_params);
  return result.toString();
}//_replaceNameByIndexPlaceholders

/**Distributes each parameter of i_params to the Object[] result or to
  a StringBuffer.
  The parameters not marked as in-text will be appended in a standard format
  to io_outtextParameters.
  @return The values of the parameters 0 to less than _maxNumberOfIndexParams.
*/
private static Object[] _distributeParams(
  final StringBuffer io_outtextParams,
  final _Param[] i_params
){
  //Copy as many params as possible to result and append all other or
  //out-text params to io_outtextParams:
  final Object[] result = new Object[_maxNumberOfIndexParams];
  if(i_params.length>0){io_outtextParams.append("  ");}
  for(int pi=0; pi<i_params.length && pi<result.length; pi++){
    final _Param param = i_params[pi];
    if(param.inText){
      result[pi] = param.value;
    }else{
      Exc.appendParam(io_outtextParams, param.name, param.value);
    }
  }
  //_println("_distributeParams-Result", result);
  return result;
}//_distributeParams

/*debug
private static void _println(final String i_name, final Object[] i_array){
  System.err.print(i_name + ": ");
  for(int i=0; i<i_array.length; i++){
    System.err.print(i_array[i] + "  ");
  }
  System.err.println();
}
*/


private static final int _noMatch = -1;

/**Tries to match a parameter in the message pattern.
  Checks if in i_messagePattern from i_startIndex there is the name
  of a parameter from io_params. If so, returns the index of the matched param.
  Otherwise returns _noMatch.
*/
private static int _indexForMatchedParam(
  final String i_messagePattern,
  final _Param[] i_params,
  final int i_startIndex
){
  final int patLen = i_messagePattern.length();
  if(i_startIndex >= patLen){return _noMatch;}

  for(int pi=0; pi<i_params.length && pi<_maxNumberOfIndexParams; pi++){ //parameter loop
    final _Param param = i_params[pi];
    final String name = param.name;
    final int namLen = name.length();
    final int endDelimI = i_startIndex + namLen;
    final char endDelim
    = endDelimI<patLen ? i_messagePattern.charAt(endDelimI) : '\0';
    if(i_messagePattern.regionMatches(i_startIndex, name, 0, namLen)
      //parameter name is matched
      && "},".indexOf(endDelim)>=0 //end delimiter is matched
    ){
      return pi;
    }
  }
  return _noMatch;
}

/**Collects the info of a named message parameter*/
private static class _Param {
  private final String name;
  private final Object value;
  private boolean inText;
  private _Param(final String i_name, final Object i_value){
    name = i_name;   value = i_value;   inText = false;
  }
  public String toString(){return name+"="+value+":"+inText;}
}

/**The maximum number of index parameters, as allowed by
  java.text.MessageFormat
*/
private static final int _maxNumberOfIndexParams = 10;

/**Returns the line separator for the actual platform.
  This is the line separator for the platform the Java virtual machine is
  running on. It is the value of the system property 'line.separator'.
*/
public static final String lineSeparator
= System.getProperty("line.separator");


//Message text access:

private static java.util.ResourceBundle _messages;
private static boolean                  _loadMessagesTried = false;

/**Tries once to load the ResourceBundle _className as message text file into
  _messages. If the loading failes, this method will leave _messages null, and
  will do nothing on further invocations.
*/
private static void _mut_tryLoadMessages(){
  if(_loadMessagesTried){return;}
  try{
    _loadMessagesTried = true;
    _messages = java.util.ResourceBundle.getBundle(_className);
  } catch(final Exception ex){ //not loadable
      ex.printStackTrace();
      _messages = null;
    }
  //end try
}//_mut_tryLoadMessages


/**Do not create objects of me!*/
private MsgText(){}

}//MsgText
